import json
import copy
from datetime import datetime
from scrapy import Spider, Request
from .simpleScraper import simpleScraper

class SimpleBot(Spider):
  name = 'simpleBot'

  def __init__(self, request):
    if type(request)==str:
        self.request = json.loads(request)
    else:
        self.request = request
    self.urls = self.request['urls']

  def start_requests(self):
    for url in self.urls:
      yield Request(url=url, callback=self.parse, cb_kwargs={'request': self.request['extract_rules']})

  def parse(self, response, request, parent_data=None, paths=None, child_path=None):
    scraper = simpleScraper(response, self.parse)
    data, follow = scraper.extract(copy.deepcopy(request))
    data['date'] = datetime.today().strftime('%Y-%m-%d-%H:%M:%S')
    data['url'] = response.request.url
    if len(follow) != 0:
      for path in follow:
        nested_request = self.find_nested_request(path, data)
        yield response.follow(url=nested_request['url'], callback=self.parse, cb_kwargs={'request': nested_request['output'], 'parent_data':data, 'child_path':path, 'paths':follow}, dont_filter=True)
    elif parent_data != None:
      data = self.add_child_to_parent(parent_data, child_path, data)
      if len(paths)==1:
        yield data
      paths.pop()
    else:
      yield data

    if "next_page" in request:
      if type(request['next_page']) == str:
        request['next_page'] = {'selector': request['next_page'], 'type':'item', 'output': 'href'}
      next_page, follow = scraper.extract({'next': request['next_page']})
      if 'next' in next_page:
        if next_page['next'] is not None:
          yield response.follow(next_page['next'], self.parse, cb_kwargs={'request': request})
  
  def find_nested_request(self, element, json):
    keys = element.split('.')
    rv = json
    for key in keys:
      if key.isdigit():
        key = int(key)
      rv = rv[key]
    return rv
  
  def nested_set(self, dic, keys, value):
    d = dic
    for key in keys[:-1]:
      if key.isdigit():
        key = int(key)
      d = d[key]
    if keys[-1] in d:
        d[keys[-1]] = value
    return dic
  
  def add_child_to_parent(self, parent, child_path, value):
    keys = child_path.split('.')
    new_parent = self.nested_set(parent, keys, value)
    return new_parent



import scrapydo
import logging
from .simpleSpider import SimpleBot

main_settings = {
    'USER_AGENT':'Mozilla/5.0 (compatible; Googlebot/2.1; +http://www.google.com/bot.html)',
    'ROBOTSTXT_OBEY': False,
    'COOKIES_ENABLED': False,
    'DEFAULT_REQUEST_HEADERS': {
        'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
        'Accept-Language': 'en'
        }
}
def special_feeds(filename):
    return {
        filename: {
        'format': 'json',
        'encoding': 'utf8',
        'store_empty': False,
        'fields': None,
        'indent': 2,
        'item_export_kwargs': {
            'export_empty_fields': True
            }
        }
    }

class ScriptRunner():
    def __init__(self, delay=1, log=False, output='data.json', custom_settings={}):
        self.delay = delay
        self.log = log
        self.output = output
        self.custom_settings = custom_settings

    def scrape(self, request):
        scrapydo.setup()
        other_settings = {'DOWNLOAD_DELAY':self.delay, 'FEEDS':special_feeds(self.output)}
        scrapydo.default_settings.update(main_settings)
        self.custom_settings.update(other_settings)
        scrapydo.default_settings.update(self.custom_settings)
        if not self.log:
            logging.basicConfig(level=logging.DEBUG)
        results = scrapydo.run_spider(SimpleBot, request=request)
        print('Massage:', 'Scraping complete.')
        return results



import copy

class simpleScraper():
    def __init__(self, response, parse=None):
        self.response = response
        self.parse = parse
        self.follow = []
        self.path = ""

    def extract(self, request):
        rules = self.standardizeRules(request)
        data = self.crawl(rules, self.response)
        return data

    def standardizeRules(self, rules_dict):
        for key in list(rules_dict):
            if type(rules_dict[key]) == str:
                rules_dict[key] = {
                    "selector": rules_dict[key],
                    "type": "item",
                    "output": "text" 
                }
            if ("selector" not in rules_dict[key]) or key=='next_page':
                del rules_dict[key]
                continue
            if 'type' not in rules_dict[key]:
                rules_dict[key]['type'] = "item"
            if "output" not in rules_dict[key]:
                rules_dict[key]['output'] = "text"
            if type(rules_dict[key]['output']) != str:
                if type(rules_dict[key]['output']) == dict:
                    rules_dict[key]['output'] = self.standardizeRules(
                        rules_dict[key]['output'])
                else:
                    rules_dict[key]['output'] = "text"
        return rules_dict

    def crawl(self, rules_dict, element, index=None):
        for key in rules_dict:
            isItem = (rules_dict[key]['type'] == 'item')
            isList = (rules_dict[key]['type'] == 'list')
            isPage = (rules_dict[key]['type'] == 'page')
            isText = (rules_dict[key]['output'] == 'text')
            isAttrib = (type(rules_dict[key]['output']) == str)
            isDict = (type(rules_dict[key]['output']) == dict)
            if isItem and isText:
                rules_dict[key] = element.css(
                    f"{rules_dict[key]['selector']}::text").get()
            elif isList and isText:
                rules_dict[key] = element.css(
                    f"{rules_dict[key]['selector']}::text").extract()
            elif (isItem and not isText) or (isList and not isText):
                if (isItem and isAttrib):
                    rules_dict[key] = element.css(
                        f"{rules_dict[key]['selector']}::attr({rules_dict[key]['output']})").get()
                elif (isList and isAttrib):
                    rules_dict[key] = element.css(
                        f"{rules_dict[key]['selector']}::attr({rules_dict[key]['output']})").extract()
                elif (isItem and isDict):
                    self.path += f" {key}"
                    parent = element.css(rules_dict[key]['selector'])
                    rules_dict[key] = self.crawl(
                        rules_dict.copy()[key]['output'], parent)
                elif (isList and isDict):
                    self.path += f" {key}"
                    parents = element.css(rules_dict[key]['selector'])
                    output = []
                    for index, parent in enumerate(parents):
                        output.append(self.crawl(copy.deepcopy(
                            rules_dict)[key]['output'], parent, index)[0])
                    rules_dict[key] = output
                else:
                    rules_dict[key] = self.response.css(
                        rules_dict[key]['selector']).extract()
            elif isPage:
                if 'follow' in rules_dict[key]:
                    page_selector = f"{rules_dict[key]['selector']}::attr({rules_dict[key]['follow']})"
                else:
                    page_selector = f"{rules_dict[key]['selector']}::attr(href)"
                page = element.css(page_selector).get()
                rules_dict[key] = {'url':page, 'output':rules_dict[key]['output']}
                if index == None:
                    index = ""
                follow_object = (f"{self.path} {index} {key}").strip().replace(" ", ".")
                self.follow.append(follow_object)
            else:
                rules_dict[key] = self.response.css(rules_dict[key]['selector']).extract()
        return (rules_dict, self.follow)
